from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtCore import QRectF, Qt
from PySide6.QtWidgets import QGraphicsSimpleTextItem

from angrmanagement.config import Conf
from angrmanagement.utils import get_label_text, is_printable

from .qgraph_object import QCachedGraphicsItem

if TYPE_CHECKING:
    from angr.knowledge_plugins.cfg.memory_data import MemoryData

    from angrmanagement.data.instance import Instance


class QMemoryDataBlock(QCachedGraphicsItem):
    ADDRESS_LABEL_OFFSET = 20
    LINEAR_INSTRUCTION_OFFSET = 120
    BYTE_AREA_SPACING = 15

    def __init__(
        self, instance: Instance, infodock, addr: int, memory_data, bytes_per_line: int = 16, parent=None
    ) -> None:
        super().__init__(parent=parent)
        self.instance = instance
        self.infodock = infodock
        self.addr = addr
        self.memory_data: MemoryData = memory_data
        self.bytes_per_line: int = bytes_per_line  # TODO: Move it to Conf

        self._addr_text = None
        self._width = None
        self._height = None

        self._bytes = []

        # widgets
        self._addr_item: QGraphicsSimpleTextItem = None
        self._label_item: QGraphicsSimpleTextItem | None = None
        self._line_items: list[
            tuple[int, QGraphicsSimpleTextItem, list[QGraphicsSimpleTextItem], list[QGraphicsSimpleTextItem]]
        ] = None

        self._init_widgets()

    #
    # Public methods
    #

    @property
    def width(self):
        return self.boundingRect().width()

    @property
    def height(self):
        return self.boundingRect().height()

    def paint(self, painter, option, widget) -> None:
        should_highlight = self.infodock.is_label_selected(self.addr)

        highlight_color = Conf.disasm_view_label_highlight_color
        if should_highlight:
            painter.setBrush(highlight_color)
            painter.setPen(highlight_color)
            painter.drawRect(0, 0, self.width, self.height)

    #
    # Event handlers
    #

    def mousePressEvent(self, event) -> None:
        if event.button() == Qt.LeftButton:
            # unselect all other labels
            self.infodock.unselect_all_labels()
            # select this label
            self.infodock.select_label(self.addr)

    #
    # Private methods
    #

    def _init_widgets(self) -> None:
        self._addr_text = f"{self.addr:08x}"
        self._bytes = []
        if self.memory_data.content:
            cnt = self.memory_data.content
            if self.memory_data.size is not None:
                cnt = cnt[: self.memory_data.size]
            self._bytes += list(cnt)

        if self.memory_data.size is not None and len(self._bytes) < self.memory_data.size:
            # load more from mapped memory
            start_address = self.memory_data.addr + len(self._bytes)
            size = self.memory_data.size - len(self._bytes)
            try:
                mem_bytes = self.instance.project.loader.memory.load(start_address, size)
            except KeyError:
                mem_bytes = b""
            self._bytes += list(mem_bytes) + ["??"] * (size - len(mem_bytes))

        # address
        self._addr_item = QGraphicsSimpleTextItem(self._addr_text, self)
        self._addr_item.setFont(Conf.disasm_font)
        self._addr_item.setBrush(Conf.disasm_view_node_address_color)

        # label
        self._init_label_item()

        # bytes
        self._init_bytes()

        self._layout_items_and_update_size()

    def _init_label_item(self) -> None:
        lbl_text = get_label_text(self.addr, self.instance.kb)
        if lbl_text:
            self._label_item = QGraphicsSimpleTextItem(lbl_text, self)
            self._label_item.setFont(Conf.code_font)
            self._label_item.setBrush(Conf.disasm_view_label_color)
        else:
            if self._label_item is not None:
                self._label_item.setParentItem(None)
                self._label_item = None

    def _init_bytes(self) -> None:
        if self._line_items:
            # remove existing items
            for line in self._line_items:
                for item in line:
                    item.setParentItem(None)
            self._line_items = None

        addr = self.addr
        i = 0
        self._line_items = []

        while i < len(self._bytes):
            byte_offset = addr % self.bytes_per_line
            end_pos = i + self.bytes_per_line if byte_offset == 0 else self.bytes_per_line - byte_offset

            all_bytes = self._bytes[i:end_pos]
            # print("... print %#x, %d bytes, byte_offset %d" % (addr, len(all_bytes), byte_offset))
            addr_item, bytes_list, character_list = self._init_line(addr, byte_offset, all_bytes)
            self._line_items.append((byte_offset, addr_item, bytes_list, character_list))

            addr += end_pos - i
            i = end_pos

    def _init_line(self, addr: int, byte_offset, all_bytes):
        # colors
        printable_byte_color = Conf.disasm_view_printable_byte_color
        printable_char_color = Conf.disasm_view_printable_character_color
        unprintable_byte_color = Conf.disasm_view_unprintable_byte_color
        unprintable_char_color = Conf.disasm_view_unprintable_character_color
        unknown_byte_color = Conf.disasm_view_unknown_byte_color
        unknown_char_color = Conf.disasm_view_unknown_character_color

        # address
        addr_text = f"{addr:08x}"
        addr_item = QGraphicsSimpleTextItem(addr_text, self)
        addr_item.setBrush(Conf.disasm_view_node_address_color)
        addr_item.setFont(Conf.disasm_font)

        # draw each byte
        bytes_list = []
        for idx, byt in enumerate(all_bytes):
            if isinstance(byt, int):
                color = printable_byte_color if is_printable(byt) else unprintable_byte_color
                o = QGraphicsSimpleTextItem(f"{byt:02x}", self)
                o.setFont(Conf.disasm_font)
                o.setBrush(color)
            else:  # str, usually because it is an unknown byte, in which case the str is "??"
                o = QGraphicsSimpleTextItem(byt, self)
                o.setBrush(unknown_byte_color)
                o.setFont(Conf.disasm_font)
            bytes_list.append(o)

            line_chars = byte_offset + idx + 1  # the number of existing characters on this line, including spaces
            if line_chars % 8 == 0 and line_chars != self.bytes_per_line:
                # print a deliminator
                o = QGraphicsSimpleTextItem("-", self)
                o.setBrush(Qt.black)
                o.setFont(Conf.disasm_font)
                bytes_list.append(o)

        # printable characters
        character_list = []
        for byt in all_bytes:
            if isinstance(byt, int):
                if is_printable(byt):
                    color = printable_char_color
                    ch = chr(byt)
                else:
                    color = unprintable_char_color
                    ch = "."
            else:
                color = unknown_char_color
                ch = "?"
            o = QGraphicsSimpleTextItem(ch, self)
            o.setBrush(color)
            o.setFont(Conf.disasm_font)
            character_list.append(o)

        return addr_item, bytes_list, character_list

    def _layout_items_and_update_size(self) -> None:
        x, y = 0, 0

        #
        # first line
        #

        # address
        self._addr_item.setPos(x, y)
        x += self._addr_item.boundingRect().width()

        # label
        if self._label_item:
            x += self.ADDRESS_LABEL_OFFSET
            self._label_item.setPos(x, y)

        #
        # the following lines: content
        #

        max_x = x
        x = 0
        y += self._addr_item.boundingRect().height()

        for byte_offset, addr_item, bytes_line, characters_line in self._line_items:
            addr_item.setPos(x, y)
            x += addr_item.boundingRect().width() + self.LINEAR_INSTRUCTION_OFFSET

            # skip byte offset
            byte_width = bytes_line[0].boundingRect().width()
            byte_spacing = byte_width // 2
            x += byte_offset * (byte_width + byte_spacing)

            all_bytes = 0
            pos = 0
            while pos < len(bytes_line):
                byte_ = bytes_line[pos]
                byte_.setPos(x, y)
                x += byte_width

                line_chars = (
                    byte_offset + all_bytes + 1
                )  # the number of existing characters on this line, including spaces
                if line_chars % 8 == 0 and line_chars != self.bytes_per_line:
                    # now we get a delimiter
                    pos += 1
                    delimiter = bytes_line[pos]
                    delimiter.setPos(x, y)

                x += byte_spacing
                pos += 1
                all_bytes += 1

            if (byte_offset + all_bytes) % self.bytes_per_line != 0:
                more_chars = self.bytes_per_line - (byte_offset + all_bytes % self.bytes_per_line)
                x += more_chars * (byte_width + byte_spacing)

            x += self.BYTE_AREA_SPACING

            # printable characters
            character_width = characters_line[0].boundingRect().width()
            x += byte_offset * character_width
            for o in characters_line:
                o.setPos(x, y)
                x += character_width

            max_x = max(x, max_x)

            # next line!
            x = 0
            y += bytes_line[0].boundingRect().height()

        self._width = max_x
        self._height = y
        self.recalculate_size()

    def _boundingRect(self):
        return QRectF(0, 0, self._width, self._height)



from __future__ import annotations

from .job import Job


class SimgrExploreJob(Job):
    def __init__(self, simgr, find=None, avoid=None, step_callback=None, until_callback=None, callback=None) -> None:
        super().__init__("Simulation manager exploring")
        self._simgr = simgr
        self._find = find
        self._avoid = avoid
        self._callback = callback
        self._step_callback = step_callback
        self._until_callback = until_callback
        self._interrupted = False

    def _run(self, inst):
        """Run the job. Runs in the worker thread."""

        def until_callback(*args, **kwargs):
            return self._interrupted or callable(self._until_callback) and self._until_callback(*args, **kwargs)

        self._simgr.explore(find=self._find, avoid=self._avoid, step_func=self._step_callback, until=until_callback)
        return self._simgr

    def finish(self, inst, result) -> None:
        super().finish(inst, result)
        self._callback(result)

    def __repr__(self) -> str:
        return f"Exploring {self._simgr!r}"

    def keyboard_interrupt(self) -> None:
        """Called from GUI thread. Worker thread will check self._interrupted periodically and exit the job early if
        needed."""
        self._interrupted = True

    @classmethod
    def create(cls, simgr, **kwargs):
        def callback(result) -> None:
            simgr.am_event(src="job_done", job="explore", result=result)

        return cls(simgr, callback=callback, **kwargs)



from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtCore import QRectF
from PySide6.QtWidgets import QGraphicsSimpleTextItem

from angrmanagement.config import Conf

from .qgraph_object import QCachedGraphicsItem

if TYPE_CHECKING:
    from angrmanagement.ui.workspace import Workspace


class QUnknownBlock(QCachedGraphicsItem):
    LINEAR_INSTRUCTION_OFFSET = 120
    DEFAULT_TEXT = "Unknown"

    def __init__(self, workspace: Workspace, addr: int, bytes_, parent=None) -> None:
        super().__init__(parent=parent)

        self.workspace = workspace
        self.addr = addr
        self.bytes = bytes_

        self._width = 0
        self._height = 0

        self._addr_text = None
        self._addr_item: QGraphicsSimpleTextItem = None
        self._byte_lines: list[QGraphicsSimpleTextItem] = None

        self._config = Conf

        self._init_widgets()

    #
    # Public methods
    #

    def paint(self, painter, option, widget) -> None:  # pylint: disable=unused-argument
        # painter.setRenderHints(
        #         QPainter.Antialiasing | QPainter.SmoothPixmapTransform)
        # painter.setFont(self._config.disasm_font)
        pass

    def _boundingRect(self):
        return QRectF(0, 0, self._width, self._height)

    #
    # Private methods
    #

    def _init_widgets(self) -> None:
        # Address
        self._addr_text = f"{self.addr:08x}"
        self._addr_item = QGraphicsSimpleTextItem(self._addr_text, self)
        self._addr_item.setBrush(Conf.disasm_view_node_address_color)
        self._addr_item.setFont(Conf.disasm_font)

        # Bytes
        self._byte_lines = []
        if self.bytes:
            line = ""
            for i, b in enumerate(self.bytes):
                line += f"{b:02x} "
                if i > 0 and (i + 1) % 16 == 0:
                    o = QGraphicsSimpleTextItem(line, self)
                    o.setFont(Conf.disasm_font)
                    o.setBrush(Conf.disasm_view_unprintable_byte_color)
                    self._byte_lines.append(o)
                    line = ""

                    if len(self._byte_lines) > 100:
                        o = QGraphicsSimpleTextItem("Remaining lines are omitted", self)
                        o.setFont(Conf.disasm_font)
                        o.setBrush(Conf.disasm_view_unprintable_byte_color)
                        self._byte_lines.append(o)
                        break

            if line:
                o = QGraphicsSimpleTextItem(line, self)
                o.setFont(Conf.disasm_font)
                o.setBrush(Conf.disasm_view_unprintable_byte_color)
                self._byte_lines.append(o)

        else:
            o = QGraphicsSimpleTextItem(QUnknownBlock.DEFAULT_TEXT, self)
            o.setBrush(Conf.disasm_view_unprintable_byte_color)
            o.setFont(Conf.disasm_font)
            self._byte_lines.append(o)

        self._layout_items_and_update_size()

    def _layout_items_and_update_size(self) -> None:
        x, y = 0, 0

        # address
        self._addr_item.setPos(x, y)

        x += self._addr_item.boundingRect().width()
        x += self.LINEAR_INSTRUCTION_OFFSET

        # lines
        max_x = x
        for line in self._byte_lines:
            line.setPos(x, y)
            y += line.boundingRect().height()
            max_x = max(max_x, line.boundingRect().width())

        self._width = max_x
        self._height = y

        self.recalculate_size()



from __future__ import annotations

import functools
from typing import TYPE_CHECKING, Any

from PySide6.QtCore import QAbstractTableModel, QSize, Qt
from PySide6.QtGui import QAction, QCursor, QFont
from PySide6.QtWidgets import QAbstractItemView, QHeaderView, QMenu, QTableView, QVBoxLayout

from angrmanagement.config import Conf
from angrmanagement.data.breakpoint import Breakpoint, BreakpointType
from angrmanagement.logic.debugger import DebuggerWatcher

from .view import InstanceView

if TYPE_CHECKING:
    import angr
    import PySide6

    from angrmanagement.data.instance import Instance
    from angrmanagement.ui.workspace import Workspace


class QStackTableModel(QAbstractTableModel):
    """
    Stack table model.
    """

    Headers = ["Offset", "Value"]
    COL_OFFSET = 0
    COL_VALUE = 1

    def __init__(self, log_widget: QStackTableWidget = None) -> None:
        super().__init__()
        self._log_widget = log_widget
        self.state: angr.SimState = None

    def rowCount(self, parent: PySide6.QtCore.QModelIndex = ...) -> int:  # pylint:disable=unused-argument
        return 0 if self.state is None else 15

    def columnCount(self, parent: PySide6.QtCore.QModelIndex = ...) -> int:  # pylint:disable=unused-argument
        return len(self.Headers)

    def headerData(
        self, section: int, orientation: PySide6.QtCore.Qt.Orientation, role: int = ...
    ) -> Any:  # pylint:disable=unused-argument
        if role != Qt.DisplayRole:
            return None
        if section < len(self.Headers):
            return self.Headers[section]
        return None

    def data(self, index: PySide6.QtCore.QModelIndex, role: int = ...) -> Any:
        if not index.isValid():
            return None
        row = index.row()
        col = index.column()
        if role == Qt.DisplayRole:
            return self._get_column_text(row, col)
        else:
            return None

    def _get_column_text(self, row, col: int) -> Any:
        width = self.state.arch.bits // 8
        offset = row * width
        mapping = {
            QStackTableModel.COL_OFFSET: lambda x: str(offset),
            QStackTableModel.COL_VALUE: lambda x: repr(self.state.stack_read(offset, width)),
        }
        func = mapping.get(col)
        if func is None:
            return None
        return func(row)


class QStackTableWidget(QTableView):
    """
    Stack table widget.
    """

    def __init__(self, stack_view) -> None:
        super().__init__()
        self.stack_view = stack_view

        hheader = self.horizontalHeader()
        hheader.setVisible(True)
        hheader.setStretchLastSection(True)

        vheader = self.verticalHeader()
        vheader.setVisible(False)
        vheader.setDefaultSectionSize(20)

        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)

        self.model: QStackTableModel = QStackTableModel(self)
        self.setModel(self.model)

        font = QFont(Conf.disasm_font)
        self.setFont(font)

        hheader.setSectionResizeMode(0, QHeaderView.ResizeToContents)

        self._dbg_manager = stack_view.instance.debugger_mgr
        self._dbg_watcher = DebuggerWatcher(self._on_debugger_state_updated, self._dbg_manager.debugger)
        self._on_debugger_state_updated()

    #
    # Events
    #

    def closeEvent(self, event) -> None:
        self._dbg_watcher.shutdown()
        super().closeEvent(event)

    def _on_debugger_state_updated(self) -> None:
        dbg = self._dbg_manager.debugger
        self.model.state = None if dbg.am_none else dbg.simstate
        self.model.layoutChanged.emit()

    def contextMenuEvent(self, arg__1: PySide6.QtGui.QContextMenuEvent) -> None:  # pylint:disable=unused-argument
        if not self.selectedIndexes():
            return

        mnu = self._get_breakpoint_submenu()
        mnu.exec_(QCursor.pos())

    def _set_breakpoint(self, bp_type: BreakpointType = BreakpointType.Execute) -> None:
        """
        Set breakpoint at current cursor.
        """
        state = self.model.state
        if state is None or not state.regs.sp.concrete:
            return

        selected = self.selectedIndexes()
        if not selected:
            return

        row = selected[0].row()
        width = state.arch.bits // 8
        offset = row * width + state.solver.eval(state.regs.sp)
        self.stack_view.instance.breakpoint_mgr.add_breakpoint(Breakpoint(bp_type, offset, width))

    def _get_breakpoint_submenu(self) -> QMenu:
        """
        Get context menu to add new breakpoints.
        """
        mnu = QMenu("Set &breakpoint", self)
        act = QAction("Break on &Execute", mnu)
        act.triggered.connect(functools.partial(self._set_breakpoint, BreakpointType.Execute))
        mnu.addAction(act)
        act = QAction("Break on &Read", mnu)
        act.triggered.connect(functools.partial(self._set_breakpoint, BreakpointType.Read))
        mnu.addAction(act)
        act = QAction("Break on &Write", mnu)
        act.triggered.connect(functools.partial(self._set_breakpoint, BreakpointType.Write))
        mnu.addAction(act)
        return mnu


class StackView(InstanceView):
    """
    Stack table view.
    """

    def __init__(self, workspace: Workspace, default_docking_position: str, instance: Instance) -> None:
        super().__init__("stack", workspace, default_docking_position, instance)

        self.base_caption = "Stack"
        self._tbl_widget: QStackTableWidget | None = None
        self._init_widgets()
        self.reload()

        self.width_hint = 500
        self.height_hint = 400
        self.updateGeometry()

    def reload(self) -> None:
        pass

    @staticmethod
    def minimumSizeHint():
        return QSize(200, 200)

    def _init_widgets(self) -> None:
        vlayout = QVBoxLayout()
        vlayout.setContentsMargins(0, 0, 0, 0)
        self._tbl_widget = QStackTableWidget(self)
        vlayout.addWidget(self._tbl_widget)
        self.setLayout(vlayout)



from __future__ import annotations

import networkx

from .job import Job


class DDGGenerationJob(Job):
    def __init__(self, addr: int) -> None:
        super().__init__("DDG generation")
        self._addr = addr

    def _run(self, inst):
        ddg = inst.project.analyses.VSA_DDG(vfg=inst.vfgs[self._addr], start_addr=self._addr)
        return ddg, networkx.relabel_nodes(ddg.graph, lambda n: n.insn_addr)

    def finish(self, inst, result) -> None:
        super().finish(inst, result)
        inst.ddgs[self._addr] = result

    def __repr__(self) -> str:
        return f"Generating VSA_DDG for function at {self._addr:#x}"



from __future__ import annotations

from .job import Job


class PrototypeFindingJob(Job):
    def __init__(self, on_finish=None) -> None:
        super().__init__(name="Function prototype finding", on_finish=on_finish)

    def _run(self, inst) -> None:
        func_count = len(inst.kb.functions)
        for i, func in enumerate(inst.kb.functions.values()):
            if func.is_simprocedure or func.is_plt:
                func.find_declaration()

            percentage = i / func_count * 100
            super()._progress_callback(percentage)

    def finish(self, inst, result) -> None:
        super().finish(inst, result)

    def __repr__(self) -> str:
        return "PrototypeFindingJob"



from __future__ import annotations

import difflib
import hashlib
import logging
import os
import re
from pathlib import Path
from typing import TYPE_CHECKING

from PySide6.QtCore import Qt
from PySide6.QtGui import QColor, QTextCharFormat, QTextCursor
from PySide6.QtWidgets import QFileDialog

from angrmanagement.data.instance import Instance
from angrmanagement.data.jobs.cfg_generation import CFGGenerationJob
from angrmanagement.data.jobs.loading import LoadBinaryJob
from angrmanagement.plugins import BasePlugin
from angrmanagement.ui.views import CodeView, DisassemblyView

from .diff_view import DiffCodeView, DiffDisassemblyView
from .function_diff import BFSFunctionDiff, FunctionDiff
from .settings_dialog import SettingsDialog

if TYPE_CHECKING:
    from angrmanagement.ui.workspace import Workspace

logger = logging.getLogger(__name__)


class PreciseDiffPlugin(BasePlugin):
    """
    A plugin for performing precisions diffs on two binaries loaded in angr-management.
    """

    DIFF_SETTINGS_CMD_EVT = 0
    LOAD_BINARY_CMD_EVT = 1
    RELOAD_BINARY_CMD_EVT = 2
    MENU_BUTTONS = ("Precise Diffing settings...", "Load binary for Precise Diffing...", "Refresh loaded diff binary")

    def __init__(self, workspace: Workspace) -> None:
        super().__init__(workspace)
        self.diff_instance: Instance | None = None
        self.current_revised_view: DisassemblyView | None = None
        self.current_revised_code: CodeView | None = None

        self.loaded_binary: Path | None = None
        self.diff_algo: FunctionDiff | None = None

        self.prefer_symbols = True
        self.resolve_strings = True
        self.resolve_insns = True
        self.use_addrs = False
        self.ignore_globals = True
        self.diff_algo_class = BFSFunctionDiff
        self.add_color = QColor(0xDDFFDD)
        self.decomp_add_color = QColor(141, 237, 141, int(0.5 * 255))
        self.del_color = QColor(0xFF7F7F)
        self.decomp_del_color = QColor(237, 141, 141, int(0.5 * 255))
        self.chg_color = QColor(0xF4ECC2)
        self.decomp_chg_color = QColor(247, 247, 138, int(0.5 * 255))

        self._differing_funcs = set()

        self._old_disass_keypress = None
        self._old_code_keypress = None

    #
    # UI Callback Handlers
    #

    def handle_click_menu(self, idx: int) -> None:
        if idx == self.DIFF_SETTINGS_CMD_EVT:
            dialog = SettingsDialog(self)
            dialog.exec()
            if dialog.updates and self.loaded_binary:
                self.syncronize_with_original_disassembly_view()
            self.ignore_globals = dialog._ignore_globals.isChecked()

        elif idx == self.LOAD_BINARY_CMD_EVT:
            params = {}
            # if a project is loaded, we use the base directory of the project
            if (
                not self.workspace.main_instance.project.am_none
                and self.workspace.main_instance.project.loader.main_object is not None
                and isinstance(self.workspace.main_instance.project.loader.main_object.binary, str)
            ):
                params["dir"] = os.path.dirname(self.workspace.main_instance.project.loader.main_object.binary)
            filepath, _ = QFileDialog.getOpenFileName(caption="Load Recompiled Object", **params)
            if not filepath:
                return

            self.load_revised_binary_from_file(Path(filepath))

        elif idx == self.RELOAD_BINARY_CMD_EVT and self.loaded_binary is not None:
            self.load_revised_binary_from_file(self.loaded_binary)

    def color_insn(self, addr: int, selected, disasm_view):
        if disasm_view != self.current_revised_view:
            return None

        if self.diff_algo is None:
            return None

        diff_value = self.diff_algo.addr_diff_value(addr)
        return self._color_map(diff_value)

    def color_func(self, func) -> QColor | None:
        return self.chg_color if func in self._differing_funcs else None

    #
    # View Construction
    #

    def _compute_differing_funcs(self):
        self._differing_funcs = set()
        base_funcs = self.workspace.main_instance.kb.functions
        base_proj = self.workspace.main_instance.project

        rev_funcs = self.diff_instance.kb.functions
        rev_proj = self.diff_instance.project

        hash_diff_pairs = []
        for func in base_funcs.values():
            if func.is_plt or func.is_syscall:
                continue

            func_key = func.addr if self.use_addrs else func.name
            if func_key not in rev_funcs:
                # TODO: do add/del highlighting
                continue

            rev_func = rev_funcs[func_key]
            base_f_hash = hashlib.md5(base_proj.loader.memory.load(func.addr, func.size)).hexdigest()
            rev_f_hash = hashlib.md5(rev_proj.loader.memory.load(rev_func.addr, rev_func.size)).hexdigest()

            if base_f_hash != rev_f_hash:
                hash_diff_pairs.append((func, rev_func))

        for base_func, rev_func in hash_diff_pairs:
            if (base_func.size != rev_func.size) or self._funcs_differ(base_func, rev_func):
                self._differing_funcs.add(base_func)

    def _funcs_differ(self, base_func, rev_func) -> bool:
        """
        We need to grab the disassembly view for each function (which may require focusing) and then compare them
        in the diffing algorithm that is more intensive than just comparing hashes.
        """
        base_disass = self.workspace.view_manager.first_view_in_category("disassembly")
        rev_disass = self.current_revised_view
        base_disass.function = base_func
        rev_disass.function = rev_func

        self.diff_algo = self.diff_algo_class(
            base_func,
            rev_func,
            disas_base=base_disass.disasm,
            disas_rev=rev_disass.disasm,
            view_base=base_disass,
            view_rev=rev_disass,
            resolve_strings=self.resolve_strings,
            prefer_symbols=self.prefer_symbols,
            resolve_insn_addrs=self.resolve_insns,
        )

        return self.diff_algo.differs

    def _color_map(self, diff_value):
        diff_map = {
            FunctionDiff.OBJ_ADDED: self.add_color,
            FunctionDiff.OBJ_DELETED: self.del_color,
            FunctionDiff.OBJ_CHANGED: self.chg_color,
            FunctionDiff.OBJ_UNMODIFIED: None,
        }

        return diff_map[diff_value]

    def color_functions_table(self):
        self._compute_differing_funcs()
        base_func_view = self.workspace.view_manager.first_view_in_category("functions")
        if base_func_view is not None:
            base_func_view.reset_cache_and_refresh()

    def color_graph_diff(self, og_disasm: DisassemblyView, new_disasm: DisassemblyView) -> None:
        try:
            base_func = og_disasm.function.am_obj
            rev_func = new_disasm.function.am_obj
        except (AttributeError, ValueError):
            return

        if base_func is None or rev_func is None:
            return

        self.diff_algo = self.diff_algo_class(
            base_func,
            rev_func,
            disas_base=og_disasm.disasm,
            disas_rev=new_disasm.disasm,
            view_base=og_disasm,
            view_rev=new_disasm,
            resolve_strings=self.resolve_strings,
            prefer_symbols=self.prefer_symbols,
            resolve_insn_addrs=self.resolve_insns,
        )
        new_disasm.redraw_current_graph()

    def color_pseudocode_diff(self, *args) -> None:  # pylint:disable=unused-argument
        og_code = self.workspace._get_or_create_view("pseudocode", CodeView)
        new_code = self.workspace._get_or_create_view("pseudocode_diff", DiffCodeView)
        try:
            base_func = og_code.codegen.text
            rev_func = new_code.codegen.text
            if self.ignore_globals:
                base_func = re.sub(r"g_[a-fA-F0-9]+", lambda m: "_" * len(m.group(0)), base_func)
                rev_func = re.sub(r"g_[a-fA-F0-9]+", lambda m: "_" * len(m.group(0)), rev_func)
        except (AttributeError, ValueError):
            return

        if base_func is None or rev_func is None:
            return

        base_lines = base_func.splitlines()
        rev_lines = rev_func.splitlines()
        diff_lines = list(difflib.ndiff(base_lines, rev_lines))
        # diff_lines = list(difflib.unified_diff(base_lines, rev_lines, lineterm=""))

        new_idx = 0
        old_idx = 0
        for line in diff_lines:
            real_line = line[2:]
            idx = None
            view = None
            color = None
            if line.startswith("+"):
                idx = rev_func.find(real_line, new_idx)
                new_idx = idx + len(real_line)
                color = self.decomp_add_color
                view = new_code
            elif line.startswith("-"):
                idx = base_func.find(real_line, old_idx)
                old_idx = idx + len(real_line)
                color = self.decomp_del_color
                view = og_code

            if view is not None and idx > -1:
                self.color_lines(view, idx, len(real_line), color)

    @staticmethod
    def color_lines(view: CodeView, start: int, length: int, color: QColor):
        # Create a QTextCursor from the QTextEdit
        cursor = QTextCursor(view.document)

        # Move the cursor to the start position and select the text range
        cursor.setPosition(start)
        cursor.movePosition(QTextCursor.Right, QTextCursor.KeepAnchor, length)

        # Create a QTextCharFormat and set the background color
        char_format = QTextCharFormat()
        char_format.setBackground(color)

        # Apply the char format to the selected text
        cursor.setCharFormat(char_format)

    def _destroy_revised_view(self) -> None:
        if self.current_revised_view:
            self.workspace.remove_view(self.current_revised_view)
            del self.current_revised_view

        self._differing_funcs = set()
        if self.diff_instance:
            del self.diff_instance

    def _create_instance_from_binary(self, file_path: Path) -> None:
        self.diff_instance.recompilation_plugin = self
        self.diff_instance.workspace = self.workspace

        job = LoadBinaryJob(file_path, on_finish=self._create_instance_from_binary_done)
        self.loaded_binary = file_path
        self.diff_instance.add_job(job)

    def _create_instance_from_binary_done(self, *args, **kwargs) -> None:  # pylint:disable=unused-argument
        job = CFGGenerationJob(on_finish=self._generate_binary_cfg_done)
        self.diff_instance.add_job(job)

    def _generate_binary_cfg_done(self, inst, cfg_info, *args, **kwargs) -> None:  # pylint:disable=unused-argument
        cfg_model, _ = cfg_info
        self.diff_instance.cfg = cfg_model
        self.revised_binary_loaded()

    def _create_revised_disassembly_view(self):
        new_disass = DiffDisassemblyView(self.workspace, "right", self.diff_instance)
        new_disass.category = "diff"
        new_disass.base_caption = "Precise Diff"
        self.current_revised_view = new_disass
        self.workspace.add_view(self.current_revised_view)
        return self.current_revised_view

    def _create_revised_code_view(self):
        new_psuedocode = DiffCodeView(self.workspace, "right", self.diff_instance, self.color_pseudocode_diff)
        new_psuedocode.category = "pseudocode_diff"
        new_psuedocode.base_caption = "Precise Diff Pseudocode"
        self.current_revised_code = new_psuedocode
        self.workspace.add_view(self.current_revised_code)
        return self.current_revised_code

    def jump_to_in_revised_view(self, func) -> None:
        self.current_revised_view.display_function(func)
        self.current_revised_view.jump_to(func.addr)

    def syncronize_with_original_disassembly_view(self) -> None:
        og_view = self.workspace._get_or_create_view("disassembly", DisassemblyView)
        if not og_view:
            return

        try:
            func_obj = og_view.current_function
        except NotImplementedError:
            return

        if func_obj is None or func_obj.am_obj is None:
            return

        og_func = func_obj.am_obj
        og_func_name = og_func.name

        lookup_sym = og_func.addr if self.use_addrs else og_func_name
        try:
            revised_func = self.diff_instance.kb.functions[lookup_sym]
        except KeyError:
            logger.warning(
                "The function %s does not exist in the diffed binary",
                hex(lookup_sym) if self.use_addrs else og_func_name,
            )
            return

        self.jump_to_in_revised_view(revised_func)
        self.color_graph_diff(og_view, self.current_revised_view)
        if not self._old_disass_keypress:
            self._old_disass_keypress = og_view.keyPressEvent
            og_view.keyPressEvent = self.stub_disass_keypress

            og_psuedo = self.workspace._get_or_create_view("pseudocode", CodeView)
            self._old_code_keypress = og_psuedo.keyPressEvent
            og_psuedo.keyPressEvent = self.stub_code_keypress

    def stub_disass_keypress(self, event):
        self._old_disass_keypress(event)
        if event.key() in {Qt.Key_Tab, Qt.Key_F5}:
            self.current_revised_view.keyPressEvent(event)
            self.color_pseudocode_diff(
                self.workspace._get_or_create_view("pseudocode", CodeView), self.current_revised_code
            )

    def stub_code_keypress(self, event):
        if event.key() == Qt.Key_Tab:
            self.current_revised_code.keyPressEvent(event)
        self._old_code_keypress(event)

    def load_revised_binary_from_file(self, file_path: Path) -> None:
        self._destroy_revised_view()
        self.diff_instance = Instance()
        self._create_instance_from_binary(file_path)

    def revised_binary_loaded(self) -> None:
        self._create_revised_disassembly_view()
        self._create_revised_code_view()
        self.syncronize_with_original_disassembly_view()
        self.workspace.view_manager.raise_view(self.current_revised_view)

        original_disass_view = self.diff_instance.workspace._get_or_create_view("disassembly", DisassemblyView)
        self.current_revised_view.sync_with_state_object(original_disass_view.sync_state)
        self.color_functions_table()



from __future__ import annotations

import logging
from collections import defaultdict
from pathlib import Path
from typing import TYPE_CHECKING

from binharness import AgentConnection, Environment, LocalEnvironment, Target

if TYPE_CHECKING:
    import angr

log = logging.getLogger(__name__)


class BhInstance:
    """BhInstance holds the state of binharness associated with a single instance."""

    agent_connections: list[AgentConnection]
    environments: list[tuple[str, Environment]]
    local_environment: Environment
    targets: defaultdict[Environment, list[Target]]

    def __init__(self):
        self.agent_connections = []
        self.local_environment = LocalEnvironment()
        self.environments = [
            ("Local", self.local_environment),
        ]
        self.targets = defaultdict(list)
        log.debug("Binharness initialized")

    def load_project(self, project: angr.Project) -> Target:
        target = Target(
            self.local_environment,
            Path(project.loader.main_object.binary),
            None,
            None,
            None,
        )
        self.targets[self.local_environment].append(target)
        log.debug("Binharness loaded local target")
        return target



from __future__ import annotations

from typing import TYPE_CHECKING

from angr.analyses.proximity_graph import (
    BaseProxiNode,
    CallProxiNode,
    FunctionProxiNode,
    IntegerProxiNode,
    StringProxiNode,
    UnknownProxiNode,
    VariableProxiNode,
)
from PySide6.QtCore import QRectF, Qt
from PySide6.QtGui import QPen
from PySide6.QtWidgets import QGraphicsSimpleTextItem

from angrmanagement.config import Conf

from .qgraph_object import QCachedGraphicsItem

if TYPE_CHECKING:
    import PySide6.QtWidgets

    from angrmanagement.ui.views.proximity_view import ProximityView


class QProximityGraphBlock(QCachedGraphicsItem):
    """
    Base Block
    """

    HORIZONTAL_PADDING = 5
    VERTICAL_PADDING = 5
    LINE_MARGIN = 3

    def __init__(self, is_selected: bool, proximity_view: ProximityView, node: BaseProxiNode) -> None:
        super().__init__()

        self._proximity_view = proximity_view
        self._workspace = self._proximity_view.workspace
        self._config = Conf

        self.selected = is_selected

        self._node = node

        #
        # Colors
        #
        self.FUNCTION_NODE_TEXT_COLOR = Conf.proximity_function_node_text_color
        self.STRING_NODE_TEXT_COLOR = Conf.proximity_string_node_text_color
        self.INTEGER_NODE_TEXT_COLOR = Conf.proximity_integer_node_text_color
        self.VARIABLE_NODE_TEXT_COLOR = Conf.proximity_variable_node_text_color
        self.UNKNOWN_NODE_TEXT_COLOR = Conf.proximity_unknown_node_text_color
        self.CALL_NODE_TEXT_COLOR = Conf.proximity_call_node_text_color
        self.CALL_NODE_TEXT_COLOR_PLT = Conf.proximity_call_node_text_color_plt
        self.CALL_NODE_TEXT_COLOR_SIMPROC = Conf.proximity_call_node_text_color_simproc

        self._init_widgets()
        self._update_size()

        self.setAcceptHoverEvents(True)

    def _init_widgets(self) -> None:
        pass

    def refresh(self) -> None:
        self._update_size()

    #
    # Event handlers
    #

    def mousePressEvent(self, event) -> None:  # pylint: disable=useless-super-delegation
        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event) -> None:
        if event.button() == Qt.LeftButton:
            self.selected = not self.selected
            self._proximity_view.redraw_graph()
            event.accept()

        super().mouseReleaseEvent(event)

    def mouseDoubleClickEvent(self, event) -> None:
        # Jump to the reference address of the node
        if event.button() == Qt.LeftButton:
            if self._node.ref_at:
                self._workspace.viz(next(iter(self._node.ref_at)))
            event.accept()
            return

        super().mouseDoubleClickEvent(event)

    def hoverEnterEvent(
        self, event: PySide6.QtWidgets.QGraphicsSceneHoverEvent
    ) -> None:  # pylint:disable=unused-argument
        self._proximity_view.hover_enter_block(self)

    def hoverLeaveEvent(
        self, event: PySide6.QtWidgets.QGraphicsSceneHoverEvent
    ) -> None:  # pylint:disable=unused-argument
        self._proximity_view.hover_leave_block(self)

    def _paint_boundary(self, painter) -> None:
        painter.setFont(Conf.symexec_font)
        normal_background = Conf.proximity_node_background_color
        selected_background = Conf.proximity_node_selected_background_color
        border_color = Conf.proximity_node_border_color

        # The node background
        if self.selected:
            painter.setBrush(selected_background)
        else:
            painter.setBrush(normal_background)
        painter.setPen(QPen(border_color, 1.5))
        painter.drawRect(0, 0, self.width, self.height)

    def paint(self, painter, option, widget) -> None:  # pylint: disable=unused-argument
        """
        Paint a state block on the scene.

        :param painter:
        :return: None
        """

        self._paint_boundary(painter)

        x = 0
        y = 0

        x += self.HORIZONTAL_PADDING
        y += self.VERTICAL_PADDING

        # text
        # text_label_x = x
        # painter.setPen(Qt.gray)
        # painter.drawText(text_label_x, y + self._config.symexec_font_ascent, "Unknown block")

    def _boundingRect(self):
        return QRectF(0, 0, self._width, self._height)

    #
    # Private methods
    #

    def _update_size(self) -> None:
        self._width = 25
        self._height = 25
        self.recalculate_size()


class QProximityGraphFunctionBlock(QProximityGraphBlock):
    """
    Function Block
    """

    def __init__(self, is_selected: bool, proximity_view: ProximityView, node: FunctionProxiNode) -> None:
        self._text = None
        self._text_item: QGraphicsSimpleTextItem = None
        super().__init__(is_selected, proximity_view, node)

    def _init_widgets(self) -> None:
        self._text = f"Function {self._node.func.name}"
        self._text_item = QGraphicsSimpleTextItem(self._text, self)
        self._text_item.setFont(Conf.symexec_font)
        self._text_item.setBrush(self.FUNCTION_NODE_TEXT_COLOR)
        self._text_item.setPos(self.HORIZONTAL_PADDING, self.VERTICAL_PADDING)

    def mouseDoubleClickEvent(self, event) -> None:
        if event.button() == Qt.LeftButton and (event.modifiers() & Qt.ControlModifier) == Qt.ControlModifier:
            # ctrl + double click to collapse a function call
            event.accept()
            self._proximity_view.collapse_function(self._node.func)
            return

        super().mouseDoubleClickEvent(event)

    def paint(self, painter, option, widget) -> None:
        self._paint_boundary(painter)

    def _update_size(self) -> None:
        width_candidates = [
            self.HORIZONTAL_PADDING * 2 + self._text_item.boundingRect().width(),
        ]

        self._width = max(width_candidates)
        self._height = self.VERTICAL_PADDING * 2 + self._text_item.boundingRect().height()

        self._width = max(30, self._width)
        self._height = max(10, self._height)
        self.recalculate_size()


class QProximityGraphCallBlock(QProximityGraphBlock):
    """
    Call Block
    """

    def __init__(self, is_selected: bool, proximity_view: ProximityView, node: CallProxiNode) -> None:
        self._func_name: str = None
        self._args: list[tuple[type, str]] = None

        self._func_name_item: QGraphicsSimpleTextItem = None
        self._left_parenthesis_item: QGraphicsSimpleTextItem = None
        self._args_list: list[QGraphicsSimpleTextItem] = None
        self._right_parenthesis_item: QGraphicsSimpleTextItem = None

        super().__init__(is_selected, proximity_view, node)

    def _init_widgets(self) -> None:
        self._node: CallProxiNode
        self._func_name = self._node.callee.name
        if self._node.args is not None:
            self._args = [self._argument_text(arg) for arg in self._node.args]
        else:
            self._args = []

        # func name
        self._func_name_item = QGraphicsSimpleTextItem(self._func_name, self)
        if self._node.callee.is_simprocedure or self._node.callee.is_plt:
            pen_color = self.CALL_NODE_TEXT_COLOR_SIMPROC
        else:
            pen_color = self.CALL_NODE_TEXT_COLOR
        self._func_name_item.setBrush(pen_color)
        self._func_name_item.setFont(Conf.symexec_font)
        self._func_name_item.setPos(self.HORIZONTAL_PADDING, self.VERTICAL_PADDING)

        x = self.HORIZONTAL_PADDING + self._func_name_item.boundingRect().width()
        y = self.VERTICAL_PADDING
        # left parenthesis
        self._left_parenthesis_item = QGraphicsSimpleTextItem("(", self)
        self._left_parenthesis_item.setBrush(pen_color)
        self._left_parenthesis_item.setFont(Conf.symexec_font)
        self._left_parenthesis_item.setPos(x, y)

        x += self._left_parenthesis_item.boundingRect().width()

        # arguments
        self._args_list = []
        for i, (type_, arg) in enumerate(self._args):
            if type_ is StringProxiNode:
                color = self.STRING_NODE_TEXT_COLOR
            elif type_ is IntegerProxiNode:
                color = self.INTEGER_NODE_TEXT_COLOR
            elif type_ is VariableProxiNode:
                color = self.VARIABLE_NODE_TEXT_COLOR
            elif type_ is UnknownProxiNode:
                color = self.UNKNOWN_NODE_TEXT_COLOR
            else:
                color = self.CALL_NODE_TEXT_COLOR
            o = QGraphicsSimpleTextItem(arg, self)
            o.setBrush(color)
            o.setFont(Conf.symexec_font)
            o.setPos(x, y)
            self._args_list.append(o)
            x += o.boundingRect().width()

            # comma
            if i != len(self._args) - 1:
                comma = QGraphicsSimpleTextItem(", ", self)
                comma.setBrush(pen_color)
                comma.setFont(Conf.symexec_font)
                comma.setPos(x, y)
                self._args_list.append(comma)
                x += comma.boundingRect().width()

        # right parenthesis
        self._right_parenthesis_item = QGraphicsSimpleTextItem(")", self)
        self._right_parenthesis_item.setBrush(pen_color)
        self._right_parenthesis_item.setFont(Conf.symexec_font)
        self._right_parenthesis_item.setPos(x, y)

    def _argument_text(self, arg) -> tuple[type, str]:  # pylint: disable=no-self-use
        if isinstance(arg, StringProxiNode):
            return StringProxiNode, '"' + arg.content.decode("utf-8").replace("\n", "\\n") + '"'
        elif isinstance(arg, IntegerProxiNode):
            return IntegerProxiNode, str(arg.value)
        elif isinstance(arg, VariableProxiNode):
            return VariableProxiNode, str(arg.name)
        elif isinstance(arg, UnknownProxiNode):
            return UnknownProxiNode, str(arg.dummy_value)
        return object, "Unknown"

    def mouseDoubleClickEvent(self, event) -> None:
        if event.button() == Qt.LeftButton and (event.modifiers() & Qt.ControlModifier) == Qt.ControlModifier:
            # ctrl + double click to expand a function call
            event.accept()
            self._proximity_view.expand_function(self._node.callee)
            return

        super().mouseDoubleClickEvent(event)

    def paint(self, painter, option, widget) -> None:
        self._paint_boundary(painter)

    def _update_size(self) -> None:
        width_candidates = [
            self.HORIZONTAL_PADDING * 2
            + self._func_name_item.boundingRect().width()
            + self._left_parenthesis_item.boundingRect().width()
            + sum(x.boundingRect().width() for x in self._args_list)
            + self._right_parenthesis_item.boundingRect().width()
        ]

        self._width = max(width_candidates)
        self._height = self.VERTICAL_PADDING * 2 + self._func_name_item.boundingRect().height()

        self._width = max(30, self._width)
        self._height = max(10, self._height)
        self.recalculate_size()


class QProximityGraphStringBlock(QProximityGraphBlock):
    """
    String Block
    """

    def __init__(self, is_selected: bool, proximity_view: ProximityView, node: StringProxiNode) -> None:
        self._text = None
        self._text_item: QGraphicsSimpleTextItem = None
        super().__init__(is_selected, proximity_view, node)

    def _init_widgets(self) -> None:
        self._text = '"' + self._node.content.decode("utf-8") + '"'

        self._text_item = QGraphicsSimpleTextItem(self._text, self)
        self._text_item.setBrush(self.STRING_NODE_TEXT_COLOR)
        self._text_item.setFont(Conf.symexec_font)
        self._text_item.setPos(self.HORIZONTAL_PADDING, self.VERTICAL_PADDING)

    def paint(self, painter, option, widget) -> None:
        self._paint_boundary(painter)

    def _update_size(self) -> None:
        width_candidates = [
            self.HORIZONTAL_PADDING * 2 + self._text_item.boundingRect().width(),
        ]

        self._width = max(width_candidates)
        self._height = self.VERTICAL_PADDING * 2 + self._text_item.boundingRect().height()

        self._width = max(30, self._width)
        self._height = max(10, self._height)
        self.recalculate_size()



from __future__ import annotations

from common import AngrManagementTestCase, ProjectOpenTestCase

from angrmanagement.logic.threads import gui_thread_schedule
from angrmanagement.ui.views import (
    BreakpointsView,
    CallExplorerView,
    CodeView,
    ConsoleView,
    DataDepView,
    DependencyView,
    DisassemblyView,
    FunctionsView,
    HexView,
    InteractionView,
    LogView,
    PatchesView,
    ProximityView,
    RegistersView,
    StackView,
    StringsView,
    SymexecView,
    TraceMapView,
    TracesView,
    TypesView,
)


class TestViewsOpen(AngrManagementTestCase):
    """Tests that all views open, without first opening a project."""

    def test_breakpoints_view(self):
        gui_thread_schedule(self.main.workspace.show_view, ("breakpoints", BreakpointsView))

    def test_call_explorer_view(self):
        gui_thread_schedule(self.main.workspace.show_view, ("call_explorer", CallExplorerView))

    def test_code_view(self):
        gui_thread_schedule(self.main.workspace.show_view, ("code", CodeView))

    def test_console_view(self):
        gui_thread_schedule(self.main.workspace.show_view, ("console", ConsoleView))

    def test_data_dep_view(self):
        gui_thread_schedule(self.main.workspace.show_view, ("data_dep", DataDepView))

    def test_dependency_view(self):
        gui_thread_schedule(self.main.workspace.show_view, ("dependency", DependencyView))

    def test_disassembly_view(self):
        gui_thread_schedule(self.main.workspace.show_view, ("disassembly", DisassemblyView))

    def test_functions_view(self):
        gui_thread_schedule(self.main.workspace.show_view, ("functions", FunctionsView))

    def test_hex_view(self):
        gui_thread_schedule(self.main.workspace.show_view, ("hex", HexView))

    def test_interaction_view(self):
        gui_thread_schedule(self.main.workspace.show_view, ("interaction", InteractionView))

    def test_log_view(self):
        gui_thread_schedule(self.main.workspace.show_view, ("log", LogView))

    def test_patches_view(self):
        gui_thread_schedule(self.main.workspace.show_view, ("patches", PatchesView))

    def test_proximity_view(self):
        gui_thread_schedule(self.main.workspace.show_view, ("proximity", ProximityView))

    def test_registers_view(self):
        gui_thread_schedule(self.main.workspace.show_view, ("registers", RegistersView))

    def test_stack_view(self):
        gui_thread_schedule(self.main.workspace.show_view, ("stack", StackView))

    def test_strings_view(self):
        gui_thread_schedule(self.main.workspace.show_view, ("strings", StringsView))

    def test_symexec_view(self):
        gui_thread_schedule(self.main.workspace.show_view, ("symexec", SymexecView))

    def test_trace_map_view(self):
        gui_thread_schedule(self.main.workspace.show_view, ("trace_map", TraceMapView))

    def test_traces_view(self):
        gui_thread_schedule(self.main.workspace.show_view, ("traces", TracesView))

    def test_types_view(self):
        gui_thread_schedule(self.main.workspace.show_view, ("types", TypesView))


class TestViewsOpenWithProject(ProjectOpenTestCase, TestViewsOpen):
    """Tests that all views open, after opening a project."""



from __future__ import annotations

try:
    import bintrace
    from bintrace.debugger_angr import get_angr_project_load_options_from_trace
except ImportError:
    bintrace = None


class Trace:
    """
    Base class for different trace formats.
    """

    @property
    def source(self) -> str:
        raise NotImplementedError

    @classmethod
    def trace_backend_enabled(cls) -> bool:
        return False

    def get_project_load_options(self):  # pylint:disable=no-self-use
        return None


class BintraceTrace(Trace):
    """
    Bintrace execution trace.
    """

    def __init__(self, trace: bintrace.Trace) -> None:
        assert BintraceTrace.trace_backend_enabled()
        self.trace: bintrace.Trace = trace

    @property
    def source(self) -> str:
        return self.trace.path

    @classmethod
    def load_trace(cls, path: str) -> BintraceTrace:
        trace = bintrace.Trace()
        trace.load_trace(path)
        return cls(trace)

    @classmethod
    def trace_backend_enabled(cls) -> bool:
        return bintrace is not None

    def get_project_load_options(self):
        return get_angr_project_load_options_from_trace(self.trace)



from __future__ import annotations

from typing import TYPE_CHECKING

import qtawesome

if TYPE_CHECKING:
    from PySide6.QtGui import QIcon

NAME_TO_QTAWESOME_NAME = {
    "about": "fa5s.info-circle",
    "command-palette": "ph.squares-four-light",
    "console-view": "mdi.console-line",
    "disassembly-graph": "fa5s.sitemap",
    "disassembly-linear": "msc.list-selection",
    "disassembly-view": "msc.symbol-constant",
    "docs": "mdi6.book-open-page-variant",
    "file": "mdi.file",
    "file-open": "mdi.folder-open",
    "functions-view": "mdi.function",
    "hex-view": "mdi.hexadecimal",
    "log-view": "mdi.message-bulleted",
    "patches-view": "mdi.sticker-outline",
    "plugins": "mdi.puzzle-edit",
    "preferences": "fa.gear",
    "pseudocode-view": "msc.json",
    "run-analysis": "mdi.arrow-right-drop-circle",
    "strings-view": "msc.symbol-string",
    "traces-view": "mdi.go-kart-track",
    "types-view": "msc.symbol-class",
}


def icon(key, *args, **kwargs) -> QIcon | None:
    qta_name = NAME_TO_QTAWESOME_NAME.get(key)
    return qtawesome.icon(qta_name, *args, **kwargs) if qta_name else None



from __future__ import annotations

from typing import TYPE_CHECKING

from PySide6.QtWidgets import QComboBox

if TYPE_CHECKING:
    from angrmanagement.data.instance import Instance


class QStateComboBox(QComboBox):
    def __init__(self, instance: Instance, allow_none: bool = True, parent=None) -> None:
        super().__init__(parent)
        self.states = instance.states
        self.allow_none = allow_none
        self._init_items()

    def _init_items(self):
        if self.allow_none:
            self.addItem("<None>", None)
        elif not self.states:
            raise ValueError("Created QStateComboBox with allow_none=False and no states available")
        for state in self.states:
            self.addItem(state.gui_data.name, state)

    @property
    def state(self):
        idx = self.currentIndex()
        if idx == -1:
            return None

        return self.itemData(idx)
